/**
* creates a new zoom behavior
*/
var zoom = d3.zoom().on("zoom", zoomHandler);

/**
* creates svg object and associated attributes
* applies the zoom behavior to svg
*/
var svg = d3.select("svg")
    .call(zoom),
width = svg.attr("width"),
height = svg.attr("height");
var textinfo='';
var textabstract='';

/**
* creates node object and associated attributes
*/
var node,
r=10,
color = d3.scaleOrdinal()
    .domain(["citing", "input", "cited"])
    .range([' #01d7c0', ' #8b90fe ', '  #a15eb2 ']),
yscale = d3.scaleOrdinal()
    .domain(["citing", "input", "cited"])
    .range([0, 250, 500]),
linetype=d3.scaleOrdinal()
    .domain(["line","dotted"])
    .range([("8,0"),("8,8")]),
toRemove;

/**
* creates link object
*/
var link;

/**
* creates a background
* creates a click functionality of the background
*/
var rect = svg.append("rect")
    .attr("x", 0)
    .attr("y", 0)
    .attr("height", height)
    .attr("width", width)
    .style("fill", 'white')
    .on('click', clickRect);

/**
* creates svg object (legend) and associated attributes
*/
var svglegend = d3.select("svg.legendsvg"),
legendposition = [65,95,125],
arrowlegendposition = [0,25],
arrowgroupnames = ["citation","self-citation"],
groupnames = ["citing","input","cited"];
    
var legend = svglegend.selectAll(".legend")
    .data(legendposition)
    .enter()
    .append("g")
    .attr("class","legend")
    .attr("transform", function(d,i) {return "translate(0," + d  + ")"; });
    
legend.append("text")
    .attr("x", 80)
    .attr("y", 0)
    .attr("dy", ".35em")
    .style("text-anchor", "start")
    .text(function(d,i){return groupnames[i]});
    
legend.append("circle")
    .attr("r", r)
    .attr("cx",30-r)
    .style("fill", color);
        
var legendarrow = svglegend.selectAll(".legendarr")
    .data(arrowlegendposition)
    .enter()
    .append("g")
    .attr("class","legendarr")
    .attr("transform", function(d) { return "translate(0," + d  + ")"; });
    
legendarrow.append("line")
    .attr("x1", 10)
    .attr("x2", 50)
    .attr("y1", 10)
    .attr("y2", 10)
    .style("stroke-dasharray",linetype)
    .style("stroke", '#999')
    .style("stroke-width", "1px")
    .style('pointer-events', 'none')
    .attr('marker-end',updateMarker('#999',this));

legendarrow.append("text")
    .attr("x", 80)
    .attr("y", 10)
    .attr("dy", ".35em")
    .style("text-anchor", "start")
    .text(function(d,i){return arrowgroupnames[i]});

/**
* creates a new simulation
* arranges the nodes according to their group on the y-axis
*/
var simulation = d3.forceSimulation()
    .force("link", d3.forceLink().id(function(d) {return d.doi;}).distance(60))
    .force("collide", d3.forceCollide(50))
    //.force("manyBody", d3.forceManyBody())
    .force("charge", d3.forceManyBody())
    .force("center", d3.forceCenter(width/2, height/2))
    //.force("xscale", d3.forceX().strength(1).x(function(d) {return xscale(parseInt(d.year))}))
    .alpha(0.1)
    //.alphaDecay(0)
    //.force("yscale", d3.forceY().strength(0.5).y(function(d) {return yscale(d.group)}));
    
/**
* creates group element
*/
var g = svg.append("g")
    .attr("class", "everything")

/**
* creates xAxis element
*/
var xAxis = d3.axisBottom()
    .tickFormat(function(d) {return d;})
    .ticks(10);;

/**
* draw xAxis
*/
var gX = svg.append("g")
        .attr("class", "axis axis--x")
        .attr("transform", "translate(0,25)")
    gX.append("text")
        .attr("y", 0)
        .attr("x", 80)
        .attr("text-anchor", "end")
        .attr("stroke", "black")
        .text("year");

/**
* loads JSON data and calls the update function
*/
d3.json("json_text.json").then(function(graph) {
    update(graph.links, graph.nodes);
})

/**
* calls update functions for xAxis, links and nodes
* adds the nodes and links to the simulation
* updates the positions of the links and nodes when the 
  state of the layout has changed (simulation has advanced by a tick)
* @param {object} nodes - nodes
* @param {object} links - links
*/
function update(links, nodes) { //THISS
    updateXAxis(nodes);
    updateLinks(links);
    updateNodes(nodes);

    simulation
        .nodes(nodes)
        .on("tick", tickHandler);
    simulation.force("link")
        .links(links);
    
    link
        .attr('marker-end', function(d) {return updateMarker("#999", d.target);})
        .style("stroke-dasharray",function(d){return self_cit(d.source,d.target)? ("8,8"): ("1,0")});
}

/**
* initializes and shows xAxis THISS
* @param {object} nodes - nodes
*/
function updateXAxis(nodes) {
    years = [];
    for (i = 0; i < nodes.length; i++) {
        years.push(parseInt((nodes[i]["year"]).split(" ")[2]));
    }

    console.log(years)
    xscale = d3.scaleLinear()
        .domain([d3.min(years)-10, d3.max(years)+10])
        .range([50, width-50])

    xAxis.scale(xscale);
    gX.call(xAxis);
}

/**
* initializes and shows links
* @param {object} links - links
*/
function updateLinks(links) {
    link = g.append("g")
        .selectAll(".link")
        .data(links)
        .enter()
        .append("line")
        .style("stroke-width", "1px")
        .style("stroke", "#999")
        .attr("class", "link");
}

/**
* initializes and shows nodes with circles and texts
* creates a new drag behavior and applies it to the circles
* creates a click functionality of the circles and texts
* @param {object} nodes - nodes
*/
function updateNodes(nodes) {
    node = g.selectAll(".node")
        .data(nodes)
        .enter()
        .append("g")
        .attr("class", "node")
        .attr('dx', function(d) {return xscale(parseInt((d.year).split(" ")[2]))})
        .attr("initial_x", function(d) {return d.dx;})
        .attr("initial_y", function(d) {return d.dy;})
        .call(d3.drag()
            .on("start", dragstarted)
            .on("drag", dragged)
            .on("end", dragended)
        );

    node.append("circle")
        .attr("class", "circle")
        .attr("r", function(d) {return 1.5*r+d.citations*0.05})
        .style("fill", function(d){ return color(d.group)})
        .on('click', clickNode);

    node.append("text")
        .attr("class", "text") 
        .style("font-size", "15px")
        .style('pointer-events', 'auto')
        .text(function (d) {
            const firstauthor = d.author[0].split(" ")
            return firstauthor[firstauthor.length-1];
        })
        .on('click', clickNode);
}

/**
* creates arrowhead and returns its url
* @param {string} color - color of arrowhead
* @param {string} target - target-node
*/
function updateMarker(color, target) {
    var radius=1.5*r+target.citations*0.05
    svg.append('defs').append('marker')//arrowhead
        .attr('id',color.replace("#", "")+radius)
        .attr('viewBox','-0 -5 10 10')
        .attr('refX',radius+9.5)
        .attr('refY',0)
        .attr('orient','auto')
        .attr('markerWidth',10)
        .attr('markerHeight',15)
        .attr('xoverflow','visible')
        .append('svg:path')
        .attr('d', 'M 0,-5 L 10 ,0 L 0,5')
        .attr('fill', color)//arrowhead color
        .style('stroke','none');
    return "url(" + color + radius + ")";
};

/**
* colors the circle and its links black and removes the previous markings
* @param {object} node - node
*/
function clickNode(node) {
    if(toRemove){
        d3.select(toRemove).selectAll(".circle").style("stroke","none")
    }
    toRemove = this.parentNode;
    d3.select(this.parentNode).selectAll(".circle").style("stroke","black")
    marklink(node)
    textfunc(node)
    resetbuttonhighlight()
    highlightbutton("overview")
    fix_nodes(node);
}

/**
* removes the markings of the circles and their links
*/
function clickRect() {
    d3.selectAll(".circle").style("stroke", "none")
    d3.selectAll(".link")
        .style("stroke", "#999")
        .attr('marker-end', function(d) {return updateMarker('#999', d.target);})
    document.getElementById('textbox').innerHTML = "Click node";
    fix_nodes(node);
}

function create_author_array(authors){
    //authorarray = authors.split(",")
    //authorarray = authorarray.map(elem =>{return elem.trim();})
    
    //return authorarray
}

function self_cit(source,target){
    //sourceauthors = create_author_array(source.author)
    //targetauthors = create_author_array(target.author)

    return source.author.some(item=>target.author.includes(item))
}
/**
* sets color of link (line and arrowhead) to black if it is directly connected to node
* and to grey otherwise
* @param {object} node - node
*/
function marklink(node){
    d3.selectAll(".link")
        .style("stroke", function(o) {
            return isLinkForNode(node, o) ? "black" : "#999";})
        .attr('marker-end', function(o) {
            return isLinkForNode(node, o) ? updateMarker('#000000', o.target) : updateMarker('#999', o.target);})
}

/**
* returns true if link is directly connected to node and false if it is not
* @param {object} node - node
* @param {object} link - link
*/
function isLinkForNode(node, link){
    return link.source.index == node.index || link.target.index == node.index;
}


/**
* returns last name of first author
* @param {string} authors - the comma-separated string of authors
*/
function firstauthor(authors){
    if (/,/.test(authors)==false){
        var firstauthor=/^.*\s+([\w\-]+)[\.\s]*$/.exec(authors)
    }
    else {
        var firstauthor=/^[\s\w\.\-]*\s([\w\-]+)[\.\s]*,.*$/.exec(authors)
    }
    return firstauthor[1]
}

/**
* outputs node info to textbox
* @param {object} node - data of current node
*/
function textfunc(node){
    textinfo="Title:" + '</br>' + node.name +
    '</br>' +'</br>'+"Author:"+ '</br>' +node.author+'</br>'+'</br>'+"Year:"+'</br>'
    +node.year+'</br>'+'</br>'+"doi:"+'</br>'+'<a href="'+node.doi+ '">'+node.doi
    +'</a>'+'</br>'+'</br>'+"Citations:"+'</br>'+node.citations;
    textabstract=node.abstract;
    document.getElementById('textbox').innerHTML = textinfo;
}


function highlightbutton(btn){
    resetbuttonhighlight();
    document.getElementById(btn).style.background="#DCDCDC";
}
function resetbuttonhighlight(){
    document.getElementById("overview").style.background='';
    document.getElementById("abstract").style.background='';
}

function displayabstract(a){
    if(textabstract=='' && textinfo==''){
        document.getElementById('textbox').innerHTML="Click node";
    }
    else{
    if (a==true){
        document.getElementById('textbox').innerHTML =textabstract;
    }
    else{
        document.getElementById('textbox').innerHTML =textinfo;
    }
    }   
}


/**
* updates the positions of the links and nodes
*/
function tickHandler() {
    link.attr("x1", function (d) {return xscale(parseInt((d.source.year).split(" ")[2]));})
        .attr("y1", function (d) {return d.source.y;})
        .attr("x2", function (d) {return xscale(parseInt((d.target.year).split(" ")[2]));})
        .attr("y2", function (d) {return d.target.y;});
    node.attr("transform", function (d) {return "translate(" + xscale(parseInt((d.year).split(" ")[2])) + ", " + d.y + ")";});
}

/**
* initializes the dragging of the node
* @param {object} node - data of current node
*/
function dragstarted(node) {
    if (!d3.event.active) 
        simulation.alphaTarget(0.3).restart()
    //node.fx = node.x;
    node.fy = node.y;
    fix_nodes(node);
}

/**
* applies the dragging to the node
* @param {object} node - data of current node
*/
function dragged(node) {
   // node.fx = d3.event.x;
    node.fy = d3.event.y;
    fix_nodes(node);
}
function fix_nodes(this_node) {
    node.each(function(d) {
      if (this_node != d) {
        d.fx = d.x;
        d.fy = d.y;
      }
    });
  }

/**
* ends the dragging and resets the position of the node
* @param {object} node - data of current node
*/
function dragended(node) {
    if (!d3.event.active) simulation.alphaTarget(.03);
    node.fx = null;
    //node.fy = null;
}

/**
* resets the positions of the nodes
*/
function resetGraph() {
    d3.selectAll(".node").each(function(d) {
        d.fx = d.initial_x;
        d.fy = d.initial_y;
    })
}

/**
* applies the transformation (zooming or dragging) to the g element
*/
function zoomHandler() {
    d3.select('g').attr("transform", d3.event.transform);
    var new_xScale = d3.event.transform.rescaleX(xscale)
    gX.call(xAxis.scale(new_xScale));
}

/**
* transforms svg so that that the zoom is reset
*/
function resetZoom() {
    d3.select('svg')
        .call(zoom.scaleTo, 1);
}

/**
* transforms svg so that it is centered
*/
function center() {
    d3.select('svg')
        .call(zoom.translateTo, 0.5 * width, 0.5 * height);
}


function savesvg(){
	var svgString = getSVGString(d3.select("svg.legendsvg").node());
	svgString2Image( svgString, 2*width, 2*height, 'png', save ); // passes Blob and filesize String to the callback

	function save( dataBlob, filesize ){
		saveAs( dataBlob, 'D3 vis exported to PNG.png' ); // FileSaver.js function
	}
};

// Below are the functions that handle actual exporting:
// getSVGString ( svgNode ) and svgString2Image( svgString, width, height, format, callback )
function getSVGString( svgNode ) {
	svgNode.setAttribute('xlink', 'http://www.w3.org/1999/xlink');
	var cssStyleText = getCSSStyles( svgNode );
	appendCSS( cssStyleText, svgNode );

	var serializer = new XMLSerializer();
	var svgString = serializer.serializeToString(svgNode);
	svgString = svgString.replace(/(\w+)?:?xlink=/g, 'xmlns:xlink='); // Fix root xlink without namespace
	svgString = svgString.replace(/NS\d+:href/g, 'xlink:href'); // Safari NS namespace fix

	return svgString;

	function getCSSStyles( parentElement ) {
		var selectorTextArr = [];

		// Add Parent element Id and Classes to the list
		selectorTextArr.push( '#'+parentElement.id );
		for (var c = 0; c < parentElement.classList.length; c++)
				if ( !contains('.'+parentElement.classList[c], selectorTextArr) )
					selectorTextArr.push( '.'+parentElement.classList[c] );

		// Add Children element Ids and Classes to the list
		var nodes = parentElement.getElementsByTagName("*");
		for (var i = 0; i < nodes.length; i++) {
			var id = nodes[i].id;
			if ( !contains('#'+id, selectorTextArr) )
				selectorTextArr.push( '#'+id );

			var classes = nodes[i].classList;
			for (var c = 0; c < classes.length; c++)
				if ( !contains('.'+classes[c], selectorTextArr) )
					selectorTextArr.push( '.'+classes[c] );
		}

		// Extract CSS Rules
		var extractedCSSText = "";
		for (var i = 0; i < document.styleSheets.length; i++) {
			var s = document.styleSheets[i];
			
			try {
			    if(!s.cssRules) continue;
			} catch( e ) {
		    		if(e.name !== 'SecurityError') throw e; // for Firefox
		    		continue;
		    	}

			var cssRules = s.cssRules;
			for (var r = 0; r < cssRules.length; r++) {
				if ( contains( cssRules[r].selectorText, selectorTextArr ) )
					extractedCSSText += cssRules[r].cssText;
			}
		}
		

		return extractedCSSText;

		function contains(str,arr) {
			return arr.indexOf( str ) === -1 ? false : true;
		}

	}

	function appendCSS( cssText, element ) {
		var styleElement = document.createElement("style");
		styleElement.setAttribute("type","text/css"); 
		styleElement.innerHTML = cssText;
		var refNode = element.hasChildNodes() ? element.children[0] : null;
		element.insertBefore( styleElement, refNode );
	}
}


function svgString2Image( svgString, width, height, format, callback ) {
	var format = format ? format : 'png';

	var imgsrc = 'data:image/svg+xml;base64,'+ btoa( unescape( encodeURIComponent( svgString ) ) ); // Convert SVG string to data URL

	var canvas = document.createElement("canvas");
	var context = canvas.getContext("2d");

	canvas.width = width;
	canvas.height = height;

	var image = new Image();
	image.onload = function() {
		context.clearRect ( 0, 0, width, height );
		context.drawImage(image, 0, 0, width, height);

		canvas.toBlob( function(blob) {
			var filesize = Math.round( blob.length/1024 ) + ' KB';
			if ( callback ) callback( blob, filesize );
		});

		
	};

	image.src = imgsrc;
}